-------------------------------------------------------------
-- (C) Michael A Smith 1993-1997                           --
-- Taken from the book Object-oriented software in Ada 95  --
--  Published by International Thomson Publishing          --
--     See http://www.brighton.ac.uk/ada95/home.htm       --
-------------------------------------------------------------


with unchecked_deallocation, sequential_io;
package body Class_pic is

  type Element is record    --
    s_index: Index;         -- The Index
    s_data : Data;          -- The Data
  end record;

  type Leaf is record       --
    left   : Subtree;       -- Possible left node
    rec    : Element;       -- Index + data
    right  : Subtree;       -- Possible right node;
  end record;

  package io is new Sequential_io( Element );

  procedure initialize( the:in out PIC ) is
  begin
    the.tree := null;    -- No storage
  end initialize;

  procedure initialize( the:in out PIC; id:in String ) is
    per : io.File_type;    -- File descriptor
    cur : Element;         -- Persistent data record element
  begin
    set_name( the, id );                  -- Name object
    io.open( per, io.IN_FILE, id );       -- Open saved state
    while not io.end_of_file( per ) loop  -- Restore saved state
      io.read( per, cur );
      add( the, cur.s_index, cur.s_data );
    end loop;
    io.close( per );
  exception                               -- Return real exception
     when others => raise Per_error;      --  as sub code
  end initialize;

  procedure finalize( the:in out PIC ) is
    per : io.File_type;    -- File descriptor
    procedure rec_finalize( the:in Subtree ) is -- Save state
    begin
      if the /= null then                   -- Subtree save as
        io.write( per, the.rec );           --  Item
        rec_finalize( the.left );           --  LHS
        rec_finalize( the.right );          --  RHS
      end if;
    end rec_finalize;
  begin
    if to_string(the.obj_id) /= "" then     -- If save state
      io.create( per, io.OUT_FILE,
                 to_string( the.obj_id ) );
      rec_finalize( the.tree );
      io.close( per );
    end if;
    release_storage( the.tree );
  exception                              -- Return real exception
     when others => raise Per_error;     --  as sub code
  end finalize;

  procedure discard( the:in out PIC ) is
  begin
    set_name( the, "" );                 -- No name
    release_storage( the.tree );         -- Release storage
  end discard;

  procedure set_name( the:in out PIC; id:in String ) is
  begin
    the.obj_id := to_bounded_string(id); -- Set object name
  end set_name;

  function  get_name( the:in PIC ) return String is
  begin
      return to_string( the.obj_id );     -- Name of object
  end get_name;

  procedure add( the:in out PIC; i:in Index; d:in Data ) is
    procedure add_s(the:in out Subtree; i:in Index; d:in Data) is
    begin
      if the = null then
        the := new Leaf'( null, Element'(i,d), null );
      else
        if i = the.rec.s_index then     -- Index all ready exists
          raise Exists;
        elsif i > the.rec.s_index then  -- Try on RHS
          add_s( the.right, i, d );
        else                            -- LHS
          add_s( the.left, i, d );
        end if;
      end if;
    end add_s;
  begin
    add_s( the.tree, i, d );
  end add;

  procedure extract(the:in out PIC; i:in Index; d:in out Data) is
    node_is : Subtree;
  begin
    node_is := find( the.tree, i );     -- Find node with iey
    d := node_is.rec.s_data;            -- return data
  end extract;

  procedure update(the:in out PIC; i:in Index; d:in out Data) is
    node_is : Subtree;
  begin
    node_is := find( the.tree, i );     -- Find node with iey
    node_is.rec.s_data := d;            -- Update data
  end update;


  function find( the:in Subtree; i:in Index) return Subtree is
  begin
    if the = null then raise Not_there; end if;
    if i = the.rec.s_index then
      return the;                           -- Found
    else
      if i > the.rec.s_index
        then return find( the.right, i );   -- Try RHS
        else return find( the.left,  i );   -- Try LHS
      end if;
    end if;
  end find;

  procedure dispose is
    new unchecked_deallocation( Leaf, Subtree );

  procedure release_storage( the:in out Subtree ) is
  begin
    if the /= null then             -- Not empty
      release_storage( the.left );  -- Free LHS
      release_storage( the.right ); -- Free RHS
      dispose( the );               -- Dispose of item
    end if;
    the := null;                    -- Subtree root null
  end release_storage;

end Class_pic;
